/***** BEGIN LICENSE BLOCK *****
   - Version: MPL 1.1/GPL 2.0/LGPL 2.1
   -
   - The contents of this file are subject to the Mozilla Public License Version
   - 1.1 (the "License"); you may not use this file except in compliance with
   - the License. You may obtain a copy of the License at
   - http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS IS" basis,
   - WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
   - for the specific language governing rights and limitations under the
   - License.
   -
   - The Original Code is "CHM Reader".
   -
   - The Initial Developer of the Original Code is Denis Remondini.
   - Portions created by the Initial Developer are Copyright (C) 2005-2006 Denis Remondini.  
   - All Rights Reserved.
   -
   - Contributor(s): Ling Li <lilingv AT gmail DOT com>
   -
   - Alternatively, the contents of this file may be used under the terms of
   - either the GNU General Public License Version 2 or later (the "GPL"), or
   - the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
   - in which case the provisions of the GPL or the LGPL are applicable instead
   - of those above. If you wish to allow use of your version of this file only
   - under the terms of either the GPL or the LGPL, and not to allow others to
   - use your version of this file under the terms of the MPL, indicate your
   - decision by deleting the provisions above and replace them with the notice
   - and other provisions required by the LGPL or the GPL. If you do not delete
   - the provisions above, a recipient may use your version of this file under
   - the terms of any one of the MPL, the GPL or the LGPL.
   -
   - 
***** END LICENSE BLOCK *****/

// Test protocol related
const kSCHEME = "chm";
const kPROTOCOL_NAME = "CHM Protocol";
const kPROTOCOL_CONTRACTID = "@mozilla.org/network/protocol;1?name=" + kSCHEME;
const kPROTOCOL_CID = Components.ID("44ec01d7-e036-41d5-baa6-1c4f6f55c7b5");

// Mozilla defined
const kSIMPLEURI_CONTRACTID = "@mozilla.org/network/simple-uri;1";
const kIOSERVICE_CONTRACTID = "@mozilla.org/network/io-service;1";
const kINPUTSTREAMPUMP_CONTRACTID = "@mozilla.org/network/input-stream-pump;1";
const kINPUTSTREAMCHANNEL_CONTRACTID = "@mozilla.org/network/input-stream-channel;1";
const kSTRINGINPUTSTREAM_CONTRACTID = "@mozilla.org/io/string-input-stream;1";
const kLOCALFILE_CONTRACTID = "@mozilla.org/file/local;1";
const kCONSOLESERVICE_CONTRACTID = "@mozilla.org/consoleservice;1";
const nsISupports = Components.interfaces.nsISupports;
const nsIIOService = Components.interfaces.nsIIOService;
const nsIProtocolHandler = Components.interfaces.nsIProtocolHandler;
const nsIURI = Components.interfaces.nsIURI;
const nsILocalFile = Components.interfaces.nsILocalFile;
const nsIInputStreamPump = Components.interfaces.nsIInputStreamPump;
const nsIChannel = Components.interfaces.nsIChannel;
const nsIInputStreamChannel = Components.interfaces.nsIInputStreamChannel;
const nsIStringInputStream = Components.interfaces.nsIStringInputStream;
const nsIConsoleService = Components.interfaces.nsIConsoleService;

const nsOK = 0;
const nsErrorUnexpected = 0x8000ffff;

// chm defined
const kCHMFILE_CONTRACTID = "@coralsoft.com/chmreader/CHMFile;1";
const ICHMFile = Components.interfaces.ICHMFile;

function Protocol()
{
}

function HHC2RDF()
{
}

Protocol.prototype =
{
  QueryInterface: function(iid)
  {
    if (!iid.equals(nsIProtocolHandler) &&
        !iid.equals(nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  },

  scheme: kSCHEME,
  defaultPort: -1,
  protocolFlags: nsIProtocolHandler.URI_NORELATIVE |
                 nsIProtocolHandler.URI_NOAUTH,
  
  allowPort: function(port, scheme)
  {
    return false;
  },

  newURI: function(spec, charset, baseURI)
  {
    //dump("spec: " + spec + "\n");
    //dump("charset: " + charset + "\n");
    //if (baseURI) dump("basespec: " + baseURI.spec + "\n");
    var uri = Components.classes[kSIMPLEURI_CONTRACTID].createInstance(nsIURI);
    if (spec.substring(0, 1) == "#") {
        var basespec = baseURI.spec;
        var pos = basespec.indexOf("#");
        uri.spec = basespec.substring(0, pos) + spec;
    }

    else if (spec.indexOf(":") > 0) {
        uri.spec = spec;
    }

    else if (spec.substring(0, 1) != '/') {
        var basespec = baseURI.spec;
        var pos = basespec.lastIndexOf("!/");
        if (pos > 0) {
            var pagepath = basespec.substring(pos + 1, basespec.lastIndexOf('/') + 1) + spec;
            if (pagepath.lastIndexOf('/') >= 1)
                pagepath = this.removeRelative(pagepath);
            uri.spec = basespec.substring(0, pos + 1) + pagepath;
        } else
            uri.spec = basespec + "!/" + spec;
    }

    return uri;
  },

  newChannel: function(aURI)
  {
    // aURI is a nsIUri, so get a string from it using .spec
    var thefile = unescape(aURI.spec);
    var pagepath = '';

    // strip away the kSCHEME: part
    thefile = thefile.substring(thefile.indexOf(":") + 1);
    if (thefile == 'null') {
        var ios = Components.classes[kIOSERVICE_CONTRACTID].getService(nsIIOService);
        var uri = ios.newURI("about:blank", null, null);
        return ios.newChannelFromURI(uri);
    }

    var pos = thefile.indexOf("!");
    if (pos > 0) {
        pagepath = thefile.substring(thefile.indexOf("!") + 1);
        thefile = thefile.substring(0, thefile.indexOf("!"));
    }

    // Load CHM File
    var localfile = Components.classes[kLOCALFILE_CONTRACTID]
                              .createInstance(nsILocalFile);
    localfile.initWithPath(thefile.substring(thefile.indexOf("://") + 3));
    var chmfile = Components.classes[kCHMFILE_CONTRACTID].createInstance(ICHMFile);
    if (chmfile.LoadCHM(localfile) != 0) {
        dump("file not found: " + localfile.path + "\n");
        // File not found
        var ios = Components.classes[kIOSERVICE_CONTRACTID].getService(nsIIOService);
        var uri = ios.newURI("about:blank", null, null);
        return ios.newChannelFromURI(uri);
    }

    // Show the default page
    if (pagepath == '') {
        var html = "<html><head>";
        html += '<meta http-equiv="Refresh" content="0;chm:file://' + escape(localfile.path) + '!' + escape(chmfile.home) + '">';
        html += "</head><body/></html>";

        var isc = Components.classes[kINPUTSTREAMCHANNEL_CONTRACTID].createInstance(nsIInputStreamChannel);
        var sis = Components.classes[kSTRINGINPUTSTREAM_CONTRACTID].createInstance(nsIStringInputStream);
        isc.contentStream = sis;
        isc.setURI(aURI);
        isc.originalURI = aURI;
        isc.contentType = "text/html";
        isc.owner = this;

        sis.setData(html, html.length);

        return isc;
    }

    if (pagepath == "/#TOPICS") {
        return this.newTopicsChannel(aURI, chmfile); 
    }

    if (pagepath == "/#INDEX") {
        return this.newIndexChannel(aURI, chmfile); 
    }

    pos = pagepath.indexOf("#");
    if (pos > 0) {
        pagepath = pagepath.substring(0, pos);
    }

    // Create the channel
    pos = pagepath.lastIndexOf(".");
    var ext = pagepath.substring(pos + 1);
    var mime = "text/html";
    switch (ext.toLowerCase()) {
    case "gif":
        mime = "image/gif";
        break;
    case "jpg":
    case "jpeg":
        mime = "image/jpeg";
        break;
    case "png":
        mime = "image/png";
        break;
    case "css":
        mime = "text/css";
        break;
    case "mht":
        mime = "message/rfc822";
    case "txt":
        mime = "text/plain";
        break;
    case "xml":
        mime = "text/xml";
        break;
    case "xhtml":
        mime = "text/xhtml";
        break;
    }

    try {
        pagepath_ui = chmfile.resolveObject(pagepath);
    } catch(e) {
        // pagepath not found
        var consoleService = Components.classes[kCONSOLESERVICE_CONTRACTID]
                                       .getService(nsIConsoleService);
        consoleService.logStringMessage("CHM Reader: pagepath not found: " + pagepath);
    }

    var isc = Components.classes[kINPUTSTREAMCHANNEL_CONTRACTID].createInstance(nsIInputStreamChannel);
    var is = chmfile.getInputStream(pagepath_ui);
    isc.contentStream = is;
    isc.setURI(aURI);
    isc.originalURI = aURI;
    isc.contentType = mime;
    isc.contentLength = pagepath_ui.length;
    isc.owner = this;
    return isc;
  },

  newTopicsChannel: function(aURI, chmfile) {
    var c = new HHC2RDF();
    var rdf = c.convert('topics', chmfile.topics);

    var isc = Components.classes[kINPUTSTREAMCHANNEL_CONTRACTID].createInstance(nsIInputStreamChannel);
    var sis = Components.classes[kSTRINGINPUTSTREAM_CONTRACTID].createInstance(nsIStringInputStream);
    isc.contentStream = sis;
    isc.setURI(aURI);
    isc.originalURI = aURI;
    isc.contentType = "application/rdf+xml";
    //isc.contentType = "text/xml";
    isc.owner = this;

    sis.setData(rdf, rdf.length);

    return isc;
  },

  newIndexChannel: function(aURI, chmfile) {
    var c = new HHC2RDF();
    var rdf = c.convert('index', chmfile.index);

    var isc = Components.classes[kINPUTSTREAMCHANNEL_CONTRACTID].createInstance(nsIInputStreamChannel);
    var sis = Components.classes[kSTRINGINPUTSTREAM_CONTRACTID].createInstance(nsIStringInputStream);
    isc.contentStream = sis;
    isc.setURI(aURI);
    isc.originalURI = aURI;
    isc.contentType = "application/rdf+xml";
    isc.contentType = "text/xml";
    isc.owner = this;

    sis.setData(rdf, rdf.length);

    return isc;
  },

    removeRelative: function(path) {
        var parts = path.split('/');
        var final = new Array();
        var final_length = 0;
        var i;
        for (i = 0; i < parts.length; i++) {
            switch(parts[i]) {
            case '.':
            case '':
                break;
            case '..':
                if (final_length > 0) final_length--;
                break;
            default:
                final[final_length++] = parts[i];
            }
        }
        return '/' + final.join('/');
    },

}

HHC2RDF.prototype =
{
    convert: function(_type, topics)
    {
        var rdf = '<?xml version="1.0" encoding="UTF-8"?>\n';
        rdf += '<RDF:RDF xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#"\n';
        rdf += '         xmlns:chm="chrome://chmreader/' + _type + 'rdf#">\n';

        var trace = new Array();
        var tree = new Object();
        tree.children = new Array();
        tree.level = 0;
        tree.about = 'chrome://chmreader/' + _type + 'root';

        var ex = /<([a-zA-Z\/]*)([^>]*)>/g;
        var aboutid = 0;
        var m;
        var parent = null;
        var current = tree;
        while (m = ex.exec(topics)) {
            switch(m[1].toLowerCase()) {
            case 'ul':
                trace.push(parent);
                parent = current;
                break;
            case '/ul':
                current = parent;
                parent = trace.pop();
                break;
            case 'object':
                if (parent) {
                    current = new Object();
                    current.level = parent.level + 1;
                    current.children = new Array();
                    parent.children.push(current);
                }
                break;
            case 'param':
                var m1;
                if (parent != null && (m1 = m[2].match(/name="(.*)" value="(.*)"/i))) {
                    switch(m1[1].toLowerCase()) {
                    case 'name':
                        current.name = m1[2].replace(/^\s*|\s*$/g, "")
                                            .replace(/&nbsp;/g, " ");
                        break;
                    case 'local':
                        if (m1[2].length > 0) {
                            aboutid += 1;
                            current.local = unescape(m1[2].replace(/^\s*|\s*$/g, ""));
                            current.about = 'chrome://chmreader/' + _type + '/' + current.local + aboutid;
                        } else {
                            aboutid += 1;
                            current.about = 'chrome://chmreader/' + _type + '/about' + aboutid;
                            current.local = '';
                        }
                        break;
                    case 'imagenumber':
                        current.image = m1[2];
                        break;
                    default:
                        dump(m1[0] + "\n");
                    }
                }
                break;
            case '/object':
                if (parent) {
                    rdf += '<RDF:Description RDF:about="' + current.about + '">\n';
                    if (current.name)
                        rdf += '<chm:name>' + current.name + '</chm:name>\n';
                    if (current.local)
                        rdf += '<chm:href>' + current.local + '</chm:href>\n';
                    if (current.image)
                        rdf += '<chm:image>' + current.image + '</chm:image>\n';
                    rdf += '</RDF:Description>\n';
                }
                break;
            }
        }

        if (_type == 'topics')
            rdf += this.dumpTree(tree, false);
        else
            rdf += this.dumpTree(tree, true);

        rdf += '</RDF:RDF>\n';

        return rdf;
    },

    dumpTree: function(tree, needsort)
    {
        var s;
        if (tree.children.length > 0) {
            if (needsort)
                tree.children.sort(this.compareChildren);
            s = '';
            if (tree.level > 0) 
                s += '<RDF:li>\n';
            s += '<RDF:Seq about="' + tree.about + '">\n';
            var i;
            for (i = 0; i < tree.children.length; i++) {
                s += this.dumpTree(tree.children[i]);
            }
            s += '</RDF:Seq>\n';
            if (tree.level > 0)
                s += '</RDF:li>\n';
        } else {
            s = '<RDF:li RDF:resource="' + tree.about + '"/>\n';
        }
        return s;
    },

    compareChildren: function(a, b) {
        if (a.name.toLowerCase() < b.name.toLowerCase()) return -1;
        if (a.name.toLowerCase() > b.name.toLowerCase()) return 1;
        return 0;
    }
}

var ProtocolFactory = new Object();

ProtocolFactory.createInstance = function (outer, iid)
{
  if (outer != null)
    throw Components.results.NS_ERROR_NO_AGGREGATION;

  if (!iid.equals(nsIProtocolHandler) &&
      !iid.equals(nsISupports))
    throw Components.results.NS_ERROR_NO_INTERFACE;

  return new Protocol();
}


/**
 * JS XPCOM component registration goop:
 *
 * We set ourselves up to observe the xpcom-startup category.  This provides
 * us with a starting point.
 */

var TestModule = new Object();

TestModule.registerSelf = function (compMgr, fileSpec, location, type)
{
  compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
  compMgr.registerFactoryLocation(kPROTOCOL_CID,
                                  kPROTOCOL_NAME,
                                  kPROTOCOL_CONTRACTID,
                                  fileSpec, 
                                  location, 
                                  type);
}

TestModule.getClassObject = function (compMgr, cid, iid)
{
  if (!cid.equals(kPROTOCOL_CID))
    throw Components.results.NS_ERROR_NO_INTERFACE;

  if (!iid.equals(Components.interfaces.nsIFactory))
    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
    
  return ProtocolFactory;
}

TestModule.canUnload = function (compMgr)
{
  return true;
}

function NSGetModule(compMgr, fileSpec)
{
  return TestModule;
}

